from .image_interpolation_1d import sample_image_1d, insert_into_image_1d
from .image_interpolation_2d import sample_image_2d, insert_into_image_2d
from .image_interpolation_3d import sample_image_3d, insert_into_image_3d
